// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013 Tianbao Yang.
//  All Rights Reserved.
//============================================================================================
// File Name: objective.cc
// implments functions for  evaluating objective
//==============================================================================================
#include "objective.h"

void  objective(Dataset& Data, long& calc_obj_time, std::vector<double>& phi_loss_error_norm, WeightVector& W, WeightVector& AvgW, 
		        uint iter, int avg, int ls, double l1s, int myclass)
{
      long startTime=get_runtime();
      W.Setsnorm();  //used in non-boost version 

      ulong num_examples = Data.size();
      phi_loss_error_norm[1] = 0.0;
      phi_loss_error_norm[3] = 0.0;
      phi_loss_error_norm[5] = W.snorm();

      if(avg){
        phi_loss_error_norm[2]= 0.0;
        phi_loss_error_norm[4]= 0.0;
        phi_loss_error_norm[6] = AvgW.snorm()/iter/iter;
      }
      if(l1s>0.0){
          phi_loss_error_norm[5] += 2*l1s*W.l1norm();
          phi_loss_error_norm[6] += 2*l1s*AvgW.l1norm()/iter;
      }



      for (ulong i=0; i < num_examples; ++i) {
          double prediction = Data.Xw(W, i);
      	  int is_class=0;
      	  if(ls<=3) is_class = (Data.y<int>(i)==myclass)?1:-1;
          
          if(ls<3){
          	double cur_loss = 1 - is_class*prediction;
          	if (cur_loss < 0.0) cur_loss = 0.0;
              phi_loss_error_norm[1] += std::pow(cur_loss, ls);
              if (cur_loss >= 1.0) phi_loss_error_norm[3] += 1.0;
          }
          else if(ls==3){
             	double cur_loss = log( 1 + exp(-is_class*prediction) );
             	phi_loss_error_norm[1] +=cur_loss;
             	if(prediction*is_class<=0.0) phi_loss_error_norm[3] +=1.0;
          }
          else if(ls==4){
            double cur_loss = std::pow(Data.y<double>(i) - prediction, 2);
          	phi_loss_error_norm[1] +=0.5*cur_loss;
          	phi_loss_error_norm[3] +=cur_loss;
          }

          if(avg){
            prediction= Data.Xw(AvgW, i)/iter;
            
            if(ls<3){
               double cur_loss = 1 - is_class*prediction;
               if (cur_loss < 0.0) cur_loss = 0.0;
               phi_loss_error_norm[2] += std::pow(cur_loss, ls);
               if (cur_loss >= 1.0) phi_loss_error_norm[4] += 1.0;
            }
            else if(ls==3){
               double cur_loss = log( 1 + exp(-is_class*prediction) );
               phi_loss_error_norm[2] +=cur_loss;
               if(prediction*is_class<=0.0) phi_loss_error_norm[4] +=1.0;
            }
            else if(ls==4){
          	 double cur_loss =std::pow(Data.y<double>(i) - prediction, 2);
          	 phi_loss_error_norm[2] +=0.5*cur_loss;
          	 phi_loss_error_norm[4] +=cur_loss; 
            }
          }
      }
      long endTime = get_runtime();
      calc_obj_time = endTime - startTime;
}

void  objective(Dataset& Data, long& calc_obj_time, std::vector<double>& loss, std::vector<double>& error, WeightVector& W, WeightVector& AvgW, 
		        int avg, int ls, int myclass, std::streambuf* osbuf)
{
	 std::ostream predf(osbuf);
    long startTime=get_runtime();
    ulong num_examples=Data.size();
    loss[0]=0; error[0]=0;
    if(avg){loss[1]=0; error[1]=0;}
    
	for (ulong i=0; i < num_examples; ++i) {
     double prediction = Data.Xw(W, i);
     if(predf.good()) predf<<prediction;
 	 int is_class=0;
 	 if(ls<=3) is_class = (Data.y<int>(i)==myclass)?1:-1;
     
     if(ls<3){
     	double cur_loss = 1 - is_class*prediction;
     	if (cur_loss < 0.0) cur_loss = 0.0;
         loss[0] += std::pow(cur_loss, ls);
         if (cur_loss >= 1.0) error[0] += 1.0;
     }
     else if(ls==3){
        	double cur_loss = log( 1 + exp(-is_class*prediction) );
        	loss[0] +=cur_loss;
        	if(prediction*is_class<=0.0) error[0] +=1.0;
     }
     else if(ls==4){
       double cur_loss = std::pow(Data.y<double>(i) - prediction, 2);
     	loss[0] +=0.5*cur_loss;
     	error[0] +=cur_loss;
     }

     if(avg){
       prediction= Data.Xw(AvgW, i);
       if(predf.good()) predf<<" "<<prediction;
       if(ls<3){
          double cur_loss = 1 - is_class*prediction;
          if (cur_loss < 0.0) cur_loss = 0.0;
          loss[1] += std::pow(cur_loss, ls);
          if (cur_loss >= 1.0) error[1] += 1.0;
       }
       else if(ls==3){
          double cur_loss = log( 1 + exp(-is_class*prediction) );
          loss[1] +=cur_loss;
          if(prediction*is_class<=0.0) error[1] +=1.0;
       }
       else if(ls==4){
     	 double cur_loss =std::pow(Data.y<double>(i) - prediction, 2);
     	 loss[1] +=0.5*cur_loss;
     	 error[1] +=cur_loss; 
       }
     }
     if(predf.good()) predf<<std::endl;
 }
 long endTime = get_runtime();
 calc_obj_time = endTime - startTime;
}




void  objective(Dataset& Data, std::vector<int> my_Class, long& calc_obj_time, std::vector<double>& phi_loss_norm, 
		        std::vector<WeightVector>&Wc, std::vector<WeightVector>& AvgWc, uint iter, int avg, int ls, double l1s)
{
        // Calculate objective value
        long startTime=get_runtime();
        //W.Setsnorm(); // handling the reduce from pointer, can be removed when using boost:mpi:reduce
        //phi_loss_norm[0...c-1] phi not used
        //phi_loss_norm[c...2*c-1] loss of W
        //phi_loss_norm[2*c...3*c-1] loss of aW
        //phi_loss_norm[3*c...4*c-1] norm of W
        //phi_loss_norm[4*c...5*c-1] norm of aW
        
        ulong num_examples = Data.size();
        uint nc = my_Class.size();
        for(uint c=0; c<nc; c++){
          //Wc[c].Setsnorm();// handling the reduce from pointer, can be removed when using boost:mpi:reduce
          phi_loss_norm[nc+c] = 0.0;
          phi_loss_norm[3*nc+c] = Wc[c].snorm();
          
          if(avg){
            phi_loss_norm[2*nc+c]= 0.0;
            phi_loss_norm[4*nc+c] = AvgWc[c].snorm()/iter/iter;
          }
          if(l1s>0.0){
            phi_loss_norm[3*nc+c] += 2*l1s*Wc[c].l1norm();
            phi_loss_norm[4*nc+c] += 2*l1s*AvgWc[c].l1norm()/iter;
          }
        }

        for (ulong i=0; i < num_examples; ++i){
           for(uint c=0; c<nc; c++){
             double prediction = Data.Xw(Wc[c], i);
         	 int is_class= (Data.y<int>(i)==my_Class[c])?1:-1;
         	 
             if(ls<3){
            	double cur_loss = 1 - is_class*prediction;
            	if (cur_loss < 0.0) cur_loss = 0.0;
                phi_loss_norm[nc+c] += std::pow(cur_loss, ls);
             }
             else if(ls==3){
               	double cur_loss = log( 1 + exp(-is_class*prediction) );
               	phi_loss_norm[nc+c] +=cur_loss;
             }

             if(avg){
               double prediction = Data.Xw(AvgWc[c], i)/iter;
               if(ls<3){
                 double cur_loss = 1 - is_class*prediction;
                 if (cur_loss < 0.0) cur_loss = 0.0;
                 phi_loss_norm[2*nc+c] += std::pow(cur_loss, ls);
               }
               else if(ls==3){
                 double cur_loss = log( 1 + exp(-is_class*prediction) );
                 phi_loss_norm[2*nc+c] +=cur_loss;
               }
             }
           }//for class
        }//for examples
        long endTime = get_runtime();
        calc_obj_time = endTime - startTime;
}




void  objective(Dataset& Data, std::vector<int> Class, long& calc_obj_time, std::vector<double>& test_error, 
		        std::vector<WeightVector>&Wc, std::vector<WeightVector>& AvgWc, int avg, int ls, std::streambuf* osbuf)
{
	   std::ostream predf(osbuf);
        // Calculate objective value
        long startTime=get_runtime();
        
        ulong num_examples = Data.size();
        uint nc = Class.size();
        for (ulong i=0; i < num_examples; ++i) {
           int pred_label_byW=0, pred_label_byaW=0;
           double pred_score_byW=-INF, pred_score_byaW=-INF; 
           for(uint c=0; c<nc; c++){
             double prediction = Data.Xw(Wc[c], i);
             if(predf.good()) predf<<prediction;
             if(prediction>pred_score_byW){
               pred_score_byW=prediction; pred_label_byW=Class[c];
             }
             if(avg){
              prediction=Data.Xw(AvgWc[c], i);
              if(predf.good()) predf<<" "<<prediction;
              if(prediction>pred_score_byaW){
            	pred_score_byaW=prediction; pred_label_byaW=Class[c];
              }
             }
             if(predf.good()) predf<<" ";
           }//for class
           if(pred_label_byW!=Data.y<int>(i)) test_error[0] +=1;
           if(avg and pred_label_byaW!=Data.y<int>(i)) test_error[1] +=1;
           if(predf.good()) predf<<std::endl;
        }//for examples
        
        long endTime = get_runtime();
        calc_obj_time = endTime - startTime;
}


void compute_obj_dual_gap(ulong n, uint nc, std::vector<double>& phi_loss_norm, std::vector<double>& total_phi_loss, 
		            double conj_val, double lambda, double l1s, uint wsize, uint count,
		            std::vector< std::vector<double> >&obj_vecc, std::vector< std::vector<double> >&dual_gap_vecc, 
		            double& max_dual_gap)
{
	            std::vector<double> obj_vec, dual_gap_vec;
	            dual_gap_vec.reserve(nc);
		        for(uint c=0; c<nc; c++){
		          total_phi_loss[c] /=n; // phi  total_phi_loss[0...nc-1] for phi
		          total_phi_loss[nc+c] /=n;  //  loss  totaL_phi_loss[nc...2*nc-1] for loss of W
		          double obj_val= phi_loss_norm[3*nc+c] * lambda / 2.0 + total_phi_loss[nc+c];
		          obj_vec.push_back(obj_val); //loss(w) + lambda/2|w|^2
	    	      double dual_gap=0;
		          if(l1s==0.0)
		    	   dual_gap=obj_val - total_phi_loss[c] + phi_loss_norm[3*nc+c] * lambda/2.0;
		          else
		           dual_gap=obj_val - total_phi_loss[c] +  conj_val * lambda;
		          if(dual_gap<0)  dual_gap=1.0e-15;
		          if(dual_gap>max_dual_gap) max_dual_gap=dual_gap;
	              dual_gap_vec.push_back(dual_gap);
	           
	             if(wsize){
	            	 if(count){
			            total_phi_loss[2*nc+c] /=n;  //  loss  totaL_phi_loss[nc...2*nc-1] for loss of W
	        	        obj_val = phi_loss_norm[4*nc+c]*lambda/2.0 + total_phi_loss[2*nc+c];
	        	        obj_vec.push_back(obj_val);}
	            	 else
	            		obj_vec.push_back(1);
	            }    	   
		       }
		       obj_vecc.push_back(obj_vec); dual_gap_vecc.push_back(dual_gap_vec);
}
